import multiprocessing
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool


class StreamExecutor(object):
    def __init__(self, **kwargs):
        _thread = kwargs.get('thread', False)
        _process = kwargs.get('process', False)

        if _thread:
            self.pool = ThreadPool(multiprocessing.cpu_count() if _thread is True else _thread)
        elif _process:
            self.pool = Pool(multiprocessing.cpu_count() if _process is True else _process)
        else:
            self.pool = Pool()

    def map(self, predicate, iterable):
        return self.pool.map(predicate, iterable)



from streamexecutor import StreamExecutor
from compatibility import _mapper, _filter, _chainer, _ranger, _islicer
import functools


class StreamType(object):
    SEQUENTIAL, PARALLEL = range(2)


class StreamMapper(object):
    THREAD, PROCESS = range(2)


class Stream(object):
    def __init__(self, *_iterable):
        self.type = StreamType.SEQUENTIAL
        self.mapping = StreamMapper.PROCESS
        self.executor = None
        self.file_ref = None
        if len(_iterable) == 1:
            if _iterable[0] is None:
                raise TypeError("Argument is None")
            self.iterable = iter(_iterable[0])
        elif len(_iterable) > 1:
            raise TypeError("Takes only one argument")
        else:
            self.iterable = iter([])

    def __iter__(self):
        return self.iterable

    def __len__(self):
        return sum(1 for item in self.iterable)

    def __getitem__(self, position):
        i = 0
        if i <= 0:
            for it in self.iterable:
                if i == position:
                    return it
                i += 1
        raise IndexError("Stream index out of range")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __del__(self):
        if self.file_ref:
            self.file_ref.close()

    def next(self):
        return next(self.iterable)

    def parallel(self, **kwargs):
        self.type = StreamType.PARALLEL
        if self.executor is None:
            self.executor = StreamExecutor(**kwargs)
        return self

    def sequential(self):
        self.type = StreamType.SEQUENTIAL
        return self

    def distinct(self):
        distincted = []
        for it in self.iterable:
            if it not in distincted:
                distincted.append(it)
        return self.__class__(distincted)

    def size(self):
        return self.__len__()

    def list(self):
        return list(self.iterable)

    def map(self, predicate):
        if self.type == StreamType.SEQUENTIAL:
            return self.__class__(_mapper(predicate, self.iterable))
        else:
            return self.__class__(self.executor.map(predicate, self.iterable))

    def foreach(self, func):
        for it in self.iterable:
            func(it)

    def skip(self, count):
        for i in _ranger(count):
            try:
                next(self.iterable)
            except StopIteration:
                return self.__class__()
        return self.__class__(self.iterable)

    def chain(self, _iterable):
        return self.__class__(_chainer(self.iterable, _iterable))

    def filter(self, predicate):
        return self.__class__(_filter(predicate, self.iterable))

    def exclude(self, predicate):
        return self.__class__((it for it in self.iterable if not predicate(it)))

    def reduce(self, predicate, initializer=None):
        if initializer is None:
            value = next(self.iterable)
        else:
            value = initializer
        for element in self.iterable:
            value = predicate(value, element)
        return value

    def peek(self, predicate):
        def _peek(iterable, fc):
            for it in iterable:
                fc(it)
                yield it
        return self.__class__(_peek(self.iterable, predicate))

    def sort(self, **kwargs):
        _cmp = kwargs.pop('cmp', None)
        _key = kwargs.pop('key', None)
        if _cmp:
            kwargs['key'] = functools.cmp_to_key(_cmp)
        if _key:
            kwargs['key'] = _key
        return self.__class__(sorted(self.iterable, **kwargs))

    def limit(self, count):
        def _limit(iterable, _count):
            for _ in _ranger(_count):
                yield next(iterable)
        return self.__class__(_limit(self.iterable, count))

    def chunk(self, chunk_size):
        def _chunker(iterable, _chunk_size):
            while True:
                chunk = list(_islicer(iterable, _chunk_size))
                if not chunk:
                    return
                yield chunk
        return self.__class__(_chunker(self.iterable, chunk_size))

    def substream(self, start, end):
        def _substream(iterable, _start, _end):
            counter = 0
            for it in iterable:
                if _start <= counter < _end:
                    yield it
                counter += 1
        if 0 <= start < end and end >= 0:
            return self.__class__(_substream(self.iterable, start, end))
        elif start == end:
            return self.__class__([])
        else:
            raise IndexError("Stream index out of range")

    def max(self):
        return max(self)

    def min(self):
        return min(self)

    def any(self, predicate):
        if predicate is None:
            iterator = iter(self.iterable)
        else:
            iterator = self.map(predicate)
        return any(iterator)

    def all(self, predicate):
        if predicate is None:
            iterator = iter(self.iterable)
        else:
            iterator = self.map(predicate)
        return all(iterator)

    def first(self, predicate=None):
        if predicate is None:
            return next(self.iterable)
        else:
            return self.filter(predicate).first()

    def last(self, predicate=None):
        if predicate is None:
            last_it = None
            for it in self.iterable:
                last_it = it
            return last_it
        else:
            return self.filter(predicate).last()

    def reverse(self):
        return self.__class__(reversed(list(self.iterable)))

    @staticmethod
    def file(*args):
        if isinstance(args[0], file):
            return Stream((line for line in args[0]))
        elif isinstance(args[0], str):
            _f = open(*args)
            stream = Stream.file(_f)
            stream.file_ref = _f
            ''' be aware that the file will never be closed '''
            return stream

    @staticmethod
    def range(*args):
        return Stream(_ranger(*args))




import itertools
from past.builtins import cmp as _cmp

_islicer = itertools.islice

try:
    _ranger = xrange
except NameError:
    _ranger = range

try:
    _filter = itertools.ifilter
except (ImportError, AttributeError):
    _filter = filter

try:
    _mapper = itertools.imap
except (ImportError, AttributeError):
    _mapper = map

try:
    _comparer = cmp
except NameError:
    _comparer = _cmp


def _chainer(*iterables):
    for it in iterables:
        for element in it:
            yield element



import unittest
from .streampy import Stream
from .compatibility import _comparer


class CreationTest(unittest.TestCase):
    def test_create_stream_without_params(self):
        s = Stream()
        self.assertEquals(0, s.size())

    def test_create_stream_with_list_params(self):
        s = Stream([])
        self.assertEquals(0, s.size())
        s = Stream([1])
        self.assertEquals(1, s.size())

    def test_create_stream_with_generator_params(self):
        s = Stream.range(1000)
        self.assertEquals(1000, s.size())

    def test_create_stream_with_bad_type(self):
        self.assertRaises(TypeError, Stream.__init__, Stream(), None)

    def test_create_stream_with_more_than_one_param(self):
        self.assertRaises(TypeError, Stream.__init__, Stream(), *([], []))


class SizeTest(unittest.TestCase):
    def test_simple_size_1(self):
        self.assertEquals(Stream.range(999).size(), 999)

    def test_simple_size_2(self):
        self.assertEquals(Stream.range(0).size(), 0)

    def test_simple_size_3(self):
        s = Stream("azertyuiop").filter(lambda x: x == 'a')
        self.assertEquals(s.size(), 1)


class FilterTest(unittest.TestCase):
    def test_simple_filter_1(self):
        s = Stream.range(10).filter(lambda x: x % 2 == 0).list()
        self.assertEquals(s, [0, 2, 4, 6, 8])
        s = Stream(['this', 'is', 'a', 'pretty', 'cat']).filter(lambda x: x != 'cat').list()
        self.assertEquals(s, ['this', 'is', 'a', 'pretty'])

    def test_simple_filter_2(self):
        s = Stream('a simple string').filter(lambda x: x == 'a').list()
        self.assertEquals(s, ['a'])

    def test_simple_filter_3(self):
        s = Stream.range(100001).filter(lambda x: x > 50000)
        self.assertEquals(50000, s.size())


class ExcludeTest(unittest.TestCase):
    def test_simple_exclude_1(self):
        s = Stream.range(10).exclude(lambda x: x % 2 == 0).list()
        self.assertEquals(s, [1, 3, 5, 7, 9])
        s = Stream(['this', 'is', 'a', 'pretty', 'cat']).exclude(lambda x: x == 'cat').list()
        self.assertEquals(s, ['this', 'is', 'a', 'pretty'])

    def test_simple_exclude_2(self):
        s = ''.join(Stream('a simple string').exclude(lambda x: x == 'a').list())
        self.assertEquals(s, ' simple string')

    def test_simple_exclude_3(self):
        s = Stream.range(100001).exclude(lambda x: x > 50000)
        self.assertEquals(50001, s.size())


class MapTest(unittest.TestCase):
    def test_simple_map_1(self):
        def add(x, y):
            return x + y
        s = Stream.range(10000).map(lambda x: x*2).reduce(add, initializer=0)
        self.assertEquals(s, 99990000)

    def test_simple_map_2(self):
        def to_dict(obj):
            return {str(obj): obj}
        s = Stream.range(10000).map(to_dict).size()
        self.assertEquals(s, 10000)

    def test_simple_map_3(self):
        def to_dict(obj):
            return {str(obj): obj}
        s = Stream([1, 2, 3]).map(to_dict).list()
        self.assertEqual(s, [{'1': 1}, {'2': 2}, {'3': 3}])


class ChainTest(unittest.TestCase):
    def test_simple_chain_1(self):
        self.assertEqual(Stream([1]).chain([2]).size(), 2)

    def test_simple_chain_2(self):
        self.assertEqual(Stream([1]).chain([2]).chain(Stream([3, 4, 5])).size(), 5)


class SortTest(unittest.TestCase):
    def test_simple_sort_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).sort(cmp=lambda x, y: _comparer(x, y)).list(), [1, 2, 3, 4, 5, 6])

    def test_simple_sort_2(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).sort(cmp=lambda x, y: _comparer(y, x)).list(), [6, 5, 4, 3, 2, 1])


class LimitTest(unittest.TestCase):
    def test_simple_limit_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).limit(2).list(), [1, 3])

    def test_simple_limit_2(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).limit(0).list(), [])

    def test_simple_limit_3(self):
        self.assertEquals(Stream("yeah baby !").limit(4).list(), ['y', 'e', 'a', 'h'])


class AnyTest(unittest.TestCase):
    def test_simple_any_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).limit(2).any(lambda x: x == 1), True)

    def test_simple_any_2(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).limit(2).any(lambda x: x == 2), False)

    def test_simple_any_3(self):
        self.assertEquals(Stream("yeah baby !").any(lambda x: 'a' < x < 'c'), True)


class AllTest(unittest.TestCase):
    def test_simple_all_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).limit(2).all(lambda x: 1 <= x <= 6), True)

    def test_simple_all_2(self):
        self.assertEquals(Stream([3, 3, 2, 5, 4, 6]).limit(2).all(lambda x: x == 3), True)

    def test_simple_all_3(self):
        self.assertEquals(Stream('yeah baby !').all(lambda x: 'a' < x < 'c'), False)


class MinTest(unittest.TestCase):
    def test_simple_all_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).min(), 1)

    def test_simple_all_2(self):
        self.assertEquals(Stream([3, 3, 2, 5, 4, 6]).min(), 2)


class MaxTest(unittest.TestCase):
    def test_simple_all_1(self):
        self.assertEquals(Stream([1, 3, 2, 5, 4, 6]).max(), 6)

    def test_simple_all_2(self):
        self.assertEquals(Stream([4, 42, 2, 5, 4, 6]).max(), 42)


class RangeTest(unittest.TestCase):
    def test_simple_range_1(self):
        self.assertEquals(Stream.range(10).list(), list(range(10)))

    def test_simple_range_2(self):
        self.assertEquals(Stream.range(42000).list(), list(range(42000)))


class FirstTest(unittest.TestCase):
    def test_simple_first_1(self):
        self.assertEquals(Stream.range(10).first(), 0)

    def test_simple_first_2(self):
        self.assertEquals(Stream.range(10).first(predicate=lambda x: x > 1), 2)


class LastTest(unittest.TestCase):
    def test_simple_last_1(self):
        self.assertEquals(Stream.range(10).last(), 9)

    def test_simple_last_2(self):
        self.assertEquals(Stream.range(10).last(predicate=lambda x: x > 1), 9)

    def test_simple_last_3(self):
        self.assertEquals(Stream.range(423).last(predicate=lambda x: 42 < x < 46), 45)


class SkipTest(unittest.TestCase):
    def test_simple_skip_1(self):
        self.assertEquals(Stream.range(10).skip(9).list(), [9])

    def test_simple_skip_2(self):
        self.assertEquals(Stream.range(10).skip(15).list(), [])


class GetItemTest(unittest.TestCase):
    def test_simple_getitem_1(self):
        self.assertEquals(Stream.range(10)[0], 0)

    def test_simple_getitem_2(self):
        self.assertEquals(Stream.range(430)[50], 50)

    def test_simple_getitem_3(self):
        self.assertRaises(IndexError, Stream.__getitem__, Stream([]), 1)


class DistinctTest(unittest.TestCase):
    def test_simple_distinct_1(self):
        self.assertEquals(Stream.range(100).chain(Stream.range(100)).distinct().list(), Stream.range(100).list())

    def test_simple_distinct_2(self):
        self.assertEquals(Stream([1, 1, 1, 1, 1, 1.0, 2]).distinct().list(), [1, 2])


class SubstreamTest(unittest.TestCase):
    def test_simple_substream_1(self):
        self.assertEqual(Stream([1, 2, 3, 4]).substream(0, 1).list(), [1])

    def test_simple_substream_2(self):
        self.assertEqual(Stream([1, 2, 3, 4]).substream(1, 1).list(), [])

    def test_simple_substream_3(self):
        self.assertEqual(Stream([1, 2, 3, 4]).substream(1, 2).list(), [2])

    def test_simple_substream_4(self):
        s = Stream.range(100)\
            .chain(Stream.range(100, 1000))\
            .substream(100, 200).list()
        self.assertEqual(s, list(range(100, 200)))


class ChunkTest(unittest.TestCase):
    def test_simple_chunk_1(self):
        self.assertEqual(Stream.range(10).chunk(2).list(), [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]])

    def test_simple_chunk_2(self):
        self.assertEqual(Stream.range(10).chunk(3).list(), [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]])

    def test_simple_chunk_3(self):
        self.assertEqual(Stream.range(1).chunk(2).list(), [[0]])


class FunctionnalTest(unittest.TestCase):
    def test_simple_1(self):
        element = Stream\
            .range(100000)\
            .filter(lambda x: x % 2 == 0)\
            .map(lambda x: str(x))\
            .map(lambda x: 'hey{0}'.format(x))\
            .first()
        self.assertEquals(element, 'hey0')

    def test_simple_2(self):
        element = Stream\
            .range(100000)\
            .filter(lambda x: x % 2 == 0)\
            .map(lambda x: str(x))\
            .map(lambda x: 'hey{0}'.format(x))\
            .limit(10)\
            .last()
        self.assertEquals(element, 'hey18')

    def test_simple_3(self):
        element = Stream.range(100000)\
            .filter(lambda x: x % 2 == 0)\
            .map(lambda x: str(x))\
            .map(lambda x: 'Hi{0}'.format(x))\
            .map(lambda x: x.upper())\
            .filter(lambda x: x.endswith('8'))\
            .limit(10)\
            .map(lambda x: x[2])\
            .skip(1)\
            .map(int)\
            .list()
        self.assertEquals(element, list(range(1, 10)))

    def test_simple_4(self):
        element = Stream(['You', 'shall', 'not', 'pass']) \
            .map(lambda x: x.upper()) \
            .exclude(lambda x: x == 'NOT')\
            .exclude(lambda x: x == 'PASS')\
            .chain(["pass"])\
            .map(lambda x: x.upper()) \
            .list()
        self.assertEquals(element, ['YOU', 'SHALL', 'PASS'])




from setuptools import setup

setup(name='streampy',
      version='1.0',
      description='Java like stream pipelines, supports parallel operations',
      url='https://github.com/tolsac/streampy.git',
      author='Camille Tolsa',
      author_email='camille.tolsa@gmail.com',
      license='MIT',
      packages=['.'],
      zip_safe=False)
